# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest

from knack.introspection import extract_full_summary_from_signature, option_descriptions, extract_args_from_signature


def op1(self, arg1, arg2=False, arg3='mydefaultvalue', **kwargs):  # pylint: disable=unused-argument
    """ This is the command description.
    :param arg3: This is an arg for a test.
    """
    pass


def op2(self, **kwargs):  # pylint: disable=unused-argument
    # This operation has no summary
    pass


class TestExtractArgs(unittest.TestCase):

    def test_extract_args_simple(self):
        arguments = dict(extract_args_from_signature(op1))

        self.assertEqual(len(arguments), 3)
        self.assertNotIn('self', arguments)
        self.assertNotIn('kwargs', arguments)

        self.assertIn('arg1', arguments)
        self.assertListEqual(arguments['arg1'].options_list, ['--arg1'])
        self.assertTrue(arguments['arg1'].options['required'])

        self.assertIn('arg2', arguments)
        self.assertListEqual(arguments['arg2'].options_list, ['--arg2'])
        self.assertFalse(arguments['arg2'].options['required'])
        self.assertEqual(arguments['arg2'].options['action'], 'store_true')

        self.assertIn('arg3', arguments)
        self.assertListEqual(arguments['arg3'].options_list, ['--arg3'])
        self.assertFalse(arguments['arg3'].options['required'])
        self.assertEqual(arguments['arg3'].options['default'], 'mydefaultvalue')
        self.assertEqual(arguments['arg3'].options['help'], 'This is an arg for a test.')

    def test_extract_args_custom_exclude(self):
        excluded_params = ['self', 'kwargs', 'arg2', 'arg3']
        arguments = dict(extract_args_from_signature(op1, excluded_params=excluded_params))
        self.assertEqual(len(arguments), 1)
        for param in excluded_params:
            self.assertNotIn(param, arguments)


class TestOptionDescriptions(unittest.TestCase):

    def test_option_desc_simple(self):
        option_descs = option_descriptions(op1)
        self.assertIn('arg3', option_descs)
        self.assertEqual(option_descs['arg3'], 'This is an arg for a test.')


class TestExtractSummary(unittest.TestCase):

    def test_extract_summary(self):
        summary = extract_full_summary_from_signature(op1)
        self.assertEqual(summary, 'This is the command description.')

    def test_extract_summary_no_summary(self):
        summary = extract_full_summary_from_signature(op2)
        self.assertEqual(summary, '')


if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


class RecordingProcessor(object):
    def process_request(self, request):
        return request

    def process_response(self, response):
        return response

    @classmethod
    def replace_header(cls, entity, header, old, new):
        cls.replace_header_fn(entity, header, lambda v: v.replace(old, new))

    @classmethod
    def replace_header_fn(cls, entity, header, replace_fn):
        try:
            header = header.lower()
            values = entity['headers'][header]
            entity['headers'][header] = [replace_fn(v) for v in values]
        except KeyError:
            pass


class LargeRequestBodyProcessor(RecordingProcessor):
    def __init__(self, max_request_body=128):
        self._max_request_body = max_request_body

    def process_request(self, request):
        if request.body and len(request.body) > self._max_request_body * 1024:
            request.body = '!!! The request body has been omitted from the recording because its ' \
                           'size {} is larger than {}KB. !!!'.format(len(request.body),
                                                                     self._max_request_body)

        return request


class LargeResponseBodyProcessor(RecordingProcessor):
    control_flag = '<CTRL-REPLACE>'

    def __init__(self, max_response_body=256):
        self._max_response_body = max_response_body

    def process_response(self, response):
        length = len(response['body']['string'] or '')
        if length > self._max_response_body * 1024:
            response['body']['string'] = \
                "!!! The response body has been omitted from the recording because it is larger " \
                "than {max} KB. It will be replaced with blank content of {length} bytes while replay. " \
                "{flag}{length}".format(max=self._max_response_body, length=length, flag=self.control_flag)

        return response


class LargeResponseBodyReplacer(RecordingProcessor):
    def process_response(self, response):
        body = response['body']['string']

        # backward compatibility. under 2.7 response body is unicode, under 3.5 response body is
        # bytes. when set the value back, the same type must be used.
        body_is_string = isinstance(body, str)

        content_in_string = (response['body']['string'] or b'').decode('utf-8')
        index = content_in_string.find(LargeResponseBodyProcessor.control_flag)

        if index > -1:
            length = int(content_in_string[index + len(LargeResponseBodyProcessor.control_flag):])
            if body_is_string:
                response['body']['string'] = '0' * length
            else:
                response['body']['string'] = bytes([0] * length)

        return response


class GeneralNameReplacer(RecordingProcessor):
    def __init__(self):
        self.names_name = []

    def register_name_pair(self, old, new):
        self.names_name.append((old, new))

    def process_request(self, request):
        for old, new in self.names_name:
            request.uri = request.uri.replace(old, new)

            if request.body:
                body = str(request.body)
                if old in body:
                    request.body = body.replace(old, new)

        return request

    def process_response(self, response):
        for old, new in self.names_name:
            if response['body']['string']:
                response['body']['string'] = response['body']['string'].replace(old, new)

            self.replace_header(response, 'location', old, new)

        return response



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import sys
import getpass

from .log import get_logger


logger = get_logger(__name__)

_INVALID_PASSWORD_MSG = 'Passwords do not match.'


class NoTTYException(Exception):
    pass


def _input(msg):
    return input(msg)


def verify_is_a_tty():
    if not sys.stdin.isatty():
        logger.debug('No tty available.')
        raise NoTTYException()


def prompt(msg, help_string=None):
    verify_is_a_tty()
    while True:
        val = _input(msg)
        if val == '?' and help_string is not None:
            print(help_string)
            continue
        return val


def prompt_int(msg, help_string=None):
    verify_is_a_tty()

    while True:
        value = _input(msg)
        if value == '?' and help_string is not None:
            print(help_string)
            continue
        try:
            return int(value)
        except ValueError:
            logger.warning('%s is not a valid number', value)


def prompt_pass(msg='Password: ', confirm=False, help_string=None):
    verify_is_a_tty()
    while True:
        password = getpass.getpass(msg)
        if password == '?' and help_string is not None:
            print(help_string)
            continue
        if confirm:
            password2 = getpass.getpass('Confirm ' + msg)
            if password != password2:
                logger.warning(_INVALID_PASSWORD_MSG)
                continue
        return password


def prompt_y_n(msg, default=None, help_string=None):
    return _prompt_bool(msg, 'y', 'n', default=default, help_string=help_string)


def prompt_t_f(msg, default=None, help_string=None):
    return _prompt_bool(msg, 't', 'f', default=default, help_string=help_string)


def _prompt_bool(msg, true_str, false_str, default=None, help_string=None):
    verify_is_a_tty()
    if default not in [None, true_str, false_str]:
        raise ValueError("Valid values for default are {}, {} or None".format(true_str, false_str))
    y = true_str.upper() if default == true_str else true_str
    n = false_str.upper() if default == false_str else false_str
    while True:
        ans = _input('{} ({}/{}): '.format(msg, y, n))
        if ans == '?' and help_string is not None:
            print(help_string)
            continue
        if ans.lower() == n.lower():
            return False
        if ans.lower() == y.lower():
            return True
        if default and not ans:
            return default == y.lower()


def prompt_choice_list(msg, a_list, default=1, help_string=None):
    """Prompt user to select from a list of possible choices.

    :param msg:A message displayed to the user before the choice list
    :type msg: str
    :param a_list:The list of choices (list of strings or list of dicts with 'name' & 'desc')
    "type a_list: list
    :param default:The default option that should be chosen if user doesn't enter a choice
    :type default: int
    :returns: The list index of the item chosen.
    """
    verify_is_a_tty()
    options = '\n'.join([' [{}] {}{}'
                         .format(i + 1,
                                 x['name'] if isinstance(x, dict) and 'name' in x else x,
                                 ' - ' + x['desc'] if isinstance(x, dict) and 'desc' in x else '')
                         for i, x in enumerate(a_list)])
    allowed_vals = list(range(1, len(a_list) + 1))
    while True:
        val = _input('{}\n{}\nPlease enter a choice [Default choice({})]: '.format(msg, options, default))
        if val == '?' and help_string is not None:
            print(help_string)
            continue
        if not val:
            val = '{}'.format(default)
        try:
            ans = int(val)
            if ans in allowed_vals:
                # array index is 0-based, user input is 1-based
                return ans - 1
            raise ValueError
        except ValueError:
            logger.warning('Valid values are %s', allowed_vals)
